from pathlib import Path
from typing import Any

import tomli_w
from imod.mf6 import Modflow6Simulation
from imod.mf6.model import Modflow6Model
from imod.msw import GridData, MetaSwapModel, Sprinkling

from primod.mapping.node_svat_mapping import NodeSvatMapping
from primod.mapping.rch_svat_mapping import RechargeSvatMapping
from primod.mapping.wel_svat_mapping import WellSvatMapping


class MetaMod:
    """Couple MetaSWAP and MODFLOW 6.

    Parameters
    ----------
    msw_model : MetaSwapModel
        The MetaSWAP model that should be coupled.
    mf6_simulation : Modflow6Simulation
        The Modflow6 simulation that should be coupled.
    mf6_rch_pkgkey: str
        Key of Modflow 6 recharge package to which MetaSWAP is coupled.
    mf6_wel_pkgkey: str or None
        Optional key of Modflow 6 well package to which MetaSWAP sprinkling is
        coupled.
    """

    _toml_name = "imod_coupler.toml"
    _modflow6_model_dir = "Modflow6"
    _metaswap_model_dir = "MetaSWAP"

    def __init__(
        self,
        msw_model: MetaSwapModel,
        mf6_simulation: Modflow6Simulation,
        mf6_rch_pkgkey: str,
        mf6_wel_pkgkey: str | None = None,
    ):
        self.msw_model = msw_model
        self.mf6_simulation = mf6_simulation
        self.mf6_rch_pkgkey = mf6_rch_pkgkey
        self.mf6_wel_pkgkey = mf6_wel_pkgkey

        self.is_sprinkling = self._check_coupler_and_sprinkling()

    def _check_coupler_and_sprinkling(self) -> bool:
        mf6_rch_pkgkey = self.mf6_rch_pkgkey
        mf6_wel_pkgkey = self.mf6_wel_pkgkey

        gwf_names = self._get_gwf_modelnames()

        # Assume only one groundwater flow model
        # FUTURE: Support multiple groundwater flow models.
        gwf_model = self.mf6_simulation[gwf_names[0]]

        if mf6_rch_pkgkey not in gwf_model.keys():
            raise ValueError(
                f"No package named {mf6_rch_pkgkey} detected in Modflow 6 model. "
                "iMOD_coupler requires a Recharge package."
            )

        sprinkling_key = self.msw_model._get_pkg_key(Sprinkling, optional_package=True)

        sprinkling_in_msw = sprinkling_key is not None
        sprinkling_in_mf6 = mf6_wel_pkgkey in gwf_model.keys()

        if sprinkling_in_msw and not sprinkling_in_mf6:
            raise ValueError(
                f"No package named {mf6_wel_pkgkey} found in Modflow 6 model, "
                "but Sprinkling package found in MetaSWAP. "
                "iMOD Coupler requires a Well Package "
                "to couple wells."
            )
        elif not sprinkling_in_msw and sprinkling_in_mf6:
            raise ValueError(
                f"Modflow 6 Well package {mf6_wel_pkgkey} specified for sprinkling, "
                "but no Sprinkling package found in MetaSWAP model."
            )
        elif sprinkling_in_msw and sprinkling_in_mf6:
            return True
        else:
            return False

    def write(
        self,
        directory: str | Path,
        modflow6_dll: str | Path,
        metaswap_dll: str | Path,
        metaswap_dll_dependency: str | Path,
        modflow6_write_kwargs: dict[str, Any] | None = None,
    ) -> None:
        """
        Write MetaSWAP and Modflow 6 model with exchange files, as well as a
        ``.toml`` file which configures the imod coupler run.

        Parameters
        ----------
        directory: str or Path
            Directory in which to write the coupled models
        modflow6_dll: str or Path
            Path to modflow6 .dll. You can obtain this library by downloading
            `the last iMOD5 release
            <https://oss.deltares.nl/web/imod/download-imod5>`_
        metaswap_dll: str or Path
            Path to metaswap .dll. You can obtain this library by downloading
            `the last iMOD5 release
            <https://oss.deltares.nl/web/imod/download-imod5>`_
        metaswap_dll_dependency: str or Path
            Directory with metaswap .dll dependencies. Directory should contain:
            [fmpich2.dll, mpich2mpi.dll, mpich2nemesis.dll, TRANSOL.dll]. You
            can obtain these by downloading `the last iMOD5 release
            <https://oss.deltares.nl/web/imod/download-imod5>`_
        modflow6_write_kwargs: dict
            Optional dictionary with keyword arguments for the writing of
            Modflow6 models. You can use this for example to turn off the
            validation at writing (``validation=False``) or to write text files
            (``binary=False``)
        """

        if modflow6_write_kwargs is None:
            modflow6_write_kwargs = {}

        # force to Path
        directory = Path(directory)
        # For some reason the Modflow 6 model has to be written first, before
        # writing the MetaSWAP model. Else we get an Access Violation Error when
        # running the coupler.
        self.mf6_simulation.write(
            directory / self._modflow6_model_dir,
            **modflow6_write_kwargs,
        )
        self.msw_model.write(directory / self._metaswap_model_dir)

        # Write exchange files
        exchange_dir = directory / "exchanges"
        exchange_dir.mkdir(mode=755, exist_ok=True)
        self.write_exchanges(exchange_dir, self.mf6_rch_pkgkey, self.mf6_wel_pkgkey)

        coupling_dict = self._get_coupling_dict(
            exchange_dir, self.mf6_rch_pkgkey, self.mf6_wel_pkgkey
        )

        self.write_toml(
            directory,
            modflow6_dll,
            metaswap_dll,
            metaswap_dll_dependency,
            coupling_dict,
        )

    def write_toml(
        self,
        directory: str | Path,
        modflow6_dll: str | Path,
        metaswap_dll: str | Path,
        metaswap_dll_dependency: str | Path,
        coupling_dict: dict[str, Any],
    ) -> None:
        """
        Write .toml file which configures the imod coupler run.

        Parameters
        ----------
        directory: str or Path
            Directory in which to write the .toml file.
        modflow6_dll: str or Path
            Path to modflow6 .dll. You can obtain this library by downloading
            `the last iMOD5 release
            <https://oss.deltares.nl/web/imod/download-imod5>`_
        metaswap_dll: str or Path
            Path to metaswap .dll. You can obtain this library by downloading
            `the last iMOD5 release
            <https://oss.deltares.nl/web/imod/download-imod5>`_
        metaswap_dll_dependency: str or Path
            Directory with metaswap .dll dependencies. Directory should contain:
            [fmpich2.dll, mpich2mpi.dll, mpich2nemesis.dll, TRANSOL.dll]. You
            can obtain these by downloading `the last iMOD5 release
            <https://oss.deltares.nl/web/imod/download-imod5>`_
        coupling_dict: dict
            Dictionary with names of coupler packages and paths to mappings.
        """
        # force to Path
        directory = Path(directory)
        directory.mkdir(parents=True, exist_ok=True)

        toml_path = directory / self._toml_name

        coupler_toml = {
            "timing": False,
            "log_level": "INFO",
            "driver_type": "metamod",
            "driver": {
                "kernels": {
                    "modflow6": {
                        "dll": str(modflow6_dll),
                        "work_dir": f".\\{self._modflow6_model_dir}",
                    },
                    "metaswap": {
                        "dll": str(metaswap_dll),
                        "work_dir": f".\\{self._metaswap_model_dir}",
                        "dll_dep_dir": str(metaswap_dll_dependency),
                    },
                },
                "coupling": [coupling_dict],
            },
        }

        with open(toml_path, "wb") as f:
            tomli_w.dump(coupler_toml, f)

    def _get_gwf_modelnames(self) -> list[str]:
        """
        Get names of gwf models in mf6 simulation
        """
        return [
            key
            for key, value in self.mf6_simulation.items()
            if isinstance(value, Modflow6Model)
        ]

    def _get_coupling_dict(
        self,
        directory: Path,
        mf6_rch_pkgkey: str,
        mf6_wel_pkgkey: str | None,
    ) -> dict[str, Any]:
        """
        Get dictionary with names of coupler packages and paths to mappings.

        Parameters
        ----------
        directory: Path
            Directory where .dxc files are written.
        mf6_rch_pkgkey: str
            Key of Modflow 6 recharge package to which MetaSWAP is coupled.
        mf6_wel_pkgkey: str
            Key of Modflow 6 well package to which MetaSWAP sprinkling is
            coupled.

        Returns
        -------
        coupling_dict: dict
            Dictionary with names of coupler packages and paths to mappings.
        """

        coupling_dict: dict[str, Any] = {}

        gwf_names = self._get_gwf_modelnames()

        # Assume only one groundwater flow model
        # FUTURE: Support multiple groundwater flow models.
        coupling_dict["mf6_model"] = gwf_names[0]

        coupling_dict[
            "mf6_msw_node_map"
        ] = f"./{directory.name}/{NodeSvatMapping._file_name}"

        coupling_dict["mf6_msw_recharge_pkg"] = mf6_rch_pkgkey
        coupling_dict[
            "mf6_msw_recharge_map"
        ] = f"./{directory.name}/{RechargeSvatMapping._file_name}"

        coupling_dict["enable_sprinkling"] = self.is_sprinkling

        if self.is_sprinkling:
            coupling_dict["mf6_msw_well_pkg"] = mf6_wel_pkgkey
            coupling_dict[
                "mf6_msw_sprinkling_map"
            ] = f"./{directory.name}/{WellSvatMapping._file_name}"

        return coupling_dict

    def write_exchanges(
        self,
        directory: str | Path,
        mf6_rch_pkgkey: str,
        mf6_wel_pkgkey: str | None,
    ) -> None:
        """
        Write exchange files (.dxc) which map MetaSWAP's svats to Modflow 6 node
        numbers, recharge ids, and well ids.

        Parameters
        ----------
        directory: str or Path
            Directory where .dxc files are written.
        mf6_rch_pkgkey: str
            Key of Modflow 6 recharge package to which MetaSWAP is coupled.
        mf6_wel_pkgkey: str
            Key of Modflow 6 well package to which MetaSWAP sprinkling is
            coupled.
        """

        gwf_names = self._get_gwf_modelnames()

        # Assume only one groundwater flow model
        # FUTURE: Support multiple groundwater flow models.
        gwf_model = self.mf6_simulation[gwf_names[0]]

        grid_data_key = [
            pkgname
            for pkgname, pkg in self.msw_model.items()
            if isinstance(pkg, GridData)
        ][0]

        dis = gwf_model[gwf_model._get_pkgkey("dis")]

        index, svat = self.msw_model[grid_data_key].generate_index_array()
        grid_mapping = NodeSvatMapping(svat, dis)
        grid_mapping.write(directory, index, svat)

        recharge = gwf_model[mf6_rch_pkgkey]

        rch_mapping = RechargeSvatMapping(svat, recharge)
        rch_mapping.write(directory, index, svat)

        if self.is_sprinkling:
            well = gwf_model[mf6_wel_pkgkey]
            well_mapping = WellSvatMapping(svat, well)
            well_mapping.write(directory, index, svat)



import numpy as np
import xarray as xr
from imod import mf6
from imod.msw.fixed_format import VariableMetaData
from primod.mapping.mappingbase import MetaModMapping


class NodeSvatMapping(MetaModMapping):
    """
    This contains the data to connect MODFLOW 6 cells (user nodes) to MetaSWAP
    svats.

    This class is responsible for the file `nodenr2svat.dxc`.

    Parameters
    ----------
    svat: array of floats (xr.DataArray)
        SVAT units. This array must have a subunit coordinate to describe
        different land uses.
    modflow_dis: mf6.StructuredDiscretization
        Modflow 6 structured discretization
    """

    _file_name = "nodenr2svat.dxc"
    _metadata_dict = {
        "mod_id": VariableMetaData(10, 1, 9999999, int),
        "free": VariableMetaData(2, None, None, str),
        "svat": VariableMetaData(10, 1, 9999999, int),
        "layer": VariableMetaData(5, 0, 9999, int),
    }

    _with_subunit = ("mod_id", "svat", "layer")
    _to_fill = ("free",)

    def __init__(
        self,
        svat: xr.DataArray,
        modflow_dis: mf6.StructuredDiscretization,
    ):
        super().__init__()
        self.dataset["svat"] = svat
        self.dataset["layer"] = xr.full_like(svat, 1)
        idomain_top_layer = modflow_dis["idomain"].sel(layer=1, drop=True)
        # Test if equal to or larger than 1, to ignore idomain == -1 as well.
        # Don't assign to self.dataset, as grid extent might differ from svat
        self.idomain_active = idomain_top_layer >= 1
        self._pkgcheck()
        self._create_mod_id()

    def _create_mod_id(self) -> None:
        """
        Create modflow indices for the recharge layer, which is where
        infiltration will take place.
        """
        self.dataset["mod_id"] = xr.full_like(
            self.dataset["svat"], fill_value=0, dtype=np.int64
        )

        n_subunit = self.dataset["subunit"].size
        n_mod = self.idomain_active.sum()

        idomain_active = self.idomain_active.to_numpy()

        # idomain does not have a subunit dimension, so tile for n_subunits
        mod_id_1d = np.tile(np.arange(1, n_mod + 1), (n_subunit, 1))

        self.dataset["mod_id"].to_numpy()[:, idomain_active] = mod_id_1d

    def _pkgcheck(self) -> None:
        # Check if active msw cell inactive in idomain
        active = self.dataset["svat"] != 0
        inactive_in_idomain = active > self.idomain_active

        if inactive_in_idomain.any():
            raise ValueError(
                "Active MetaSWAP cell detected in inactive cell in Modflow6 idomain"
            )



import copy
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import Any

import geopandas as gpd
import imod
import numpy as np
import pandas as pd
import ribasim
import tomli_w
import xarray as xr
from imod.mf6 import Drainage, GroundwaterFlowModel, Modflow6Simulation, River
from imod.msw import GridData, MetaSwapModel, Sprinkling

from primod.mapping.node_svat_mapping import NodeSvatMapping
from primod.mapping.rch_svat_mapping import RechargeSvatMapping
from primod.mapping.wel_svat_mapping import WellSvatMapping


@dataclass
class DriverCoupling:
    """A dataclass representing one coupling scenario for the RibaMetaMod driver.

    Attributes
    ----------
    mf6_model : str
        The model of the driver.
    mf6_active_river_packages : list of str
        A list of active river packages.
    mf6_passive_river_packages : list of str
        A list of passive river packages.
    mf6_active_drainage_packages : list of str
        A list of active drainage packages.
    mf6_passive_drainage_packages : list of str
        A list of passive drainage packages.
    """

    mf6_model: str
    mf6_active_river_packages: list[str] = field(default_factory=list)
    mf6_passive_river_packages: list[str] = field(default_factory=list)
    mf6_active_drainage_packages: list[str] = field(default_factory=list)
    mf6_passive_drainage_packages: list[str] = field(default_factory=list)


class RibaMetaMod:
    """Couple Ribasim, MetaSWAP and MODFLOW 6.
    Parameters
    ----------
    ribasim_model : ribasim.model
        The Ribasim model that should be coupled.
    msw_model : MetaSwapModel
        The MetaSWAP model that should be coupled.
    mf6_simulation : Modflow6Simulation
        The Modflow6 simulation that should be coupled.
    coupling_list: list of DriverCoupling
        One entry per MODFLOW 6 model that should be coupled
    """

    _toml_name = "imod_coupler.toml"
    _ribasim_model_dir = "ribasim"
    _modflow6_model_dir = "modflow6"
    _metaswap_model_dir = "metaswap"

    def __init__(
        self,
        ribasim_model: ribasim.Model,
        msw_model: MetaSwapModel,
        mf6_simulation: Modflow6Simulation,
        basin_definition: gpd.GeoDataFrame,
        coupling_list: list[DriverCoupling],
        mf6_rch_pkgkey: str,
        mf6_wel_pkgkey: str | None = None,
    ):
        self.ribasim_model = ribasim_model
        self.mf6_simulation = mf6_simulation
        self.msw_model = msw_model
        self.coupling_list = coupling_list
        self.mf6_rch_pkgkey = mf6_rch_pkgkey
        self.mf6_wel_pkgkey = mf6_wel_pkgkey
        self.is_sprinkling = self._check_coupler_and_sprinkling()

        if "node_id" not in basin_definition.columns:
            raise ValueError('Basin definition must contain "node_id" column')
        self.basin_definition = basin_definition

    def _check_coupler_and_sprinkling(self) -> bool:
        mf6_rch_pkgkey = self.mf6_rch_pkgkey
        mf6_wel_pkgkey = self.mf6_wel_pkgkey

        gwf_names = self._get_gwf_modelnames()

        # Assume only one groundwater flow model
        # FUTURE: Support multiple groundwater flow models.
        gwf_model = self.mf6_simulation[gwf_names[0]]

        if mf6_rch_pkgkey not in gwf_model.keys():
            raise ValueError(
                f"No package named {mf6_rch_pkgkey} detected in Modflow 6 model. "
                "iMOD_coupler requires a Recharge package."
            )

        sprinkling_key = self.msw_model._get_pkg_key(Sprinkling, optional_package=True)

        sprinkling_in_msw = sprinkling_key is not None
        sprinkling_in_mf6 = mf6_wel_pkgkey in gwf_model.keys()

        if sprinkling_in_msw and not sprinkling_in_mf6:
            raise ValueError(
                f"No package named {mf6_wel_pkgkey} found in Modflow 6 model, "
                "but Sprinkling package found in MetaSWAP. "
                "iMOD Coupler requires a Well Package "
                "to couple wells."
            )
        elif not sprinkling_in_msw and sprinkling_in_mf6:
            raise ValueError(
                f"Modflow 6 Well package {mf6_wel_pkgkey} specified for sprinkling, "
                "but no Sprinkling package found in MetaSWAP model."
            )
        elif sprinkling_in_msw and sprinkling_in_mf6:
            return True
        else:
            return False

    def _get_gwf_modelnames(self) -> list[str]:
        """
        Get names of gwf models in mf6 simulation
        """
        return [
            key
            for key, value in self.mf6_simulation.items()
            if isinstance(value, GroundwaterFlowModel)
        ]

    def write(
        self,
        directory: str | Path,
        modflow6_dll: str | Path,
        ribasim_dll: str | Path,
        ribasim_dll_dependency: str | Path,
        metaswap_dll: str | Path,
        metaswap_dll_dependency: str | Path,
        modflow6_write_kwargs: dict[str, Any] | None = None,
    ) -> None:
        """
        Write Ribasim, MetaSWAP and Modflow 6 model with exchange files, as well as a
        ``.toml`` file which configures the iMOD Coupler run.

        Parameters
        ----------
        directory: str or Path
            Directory in which to write the coupled models
        modflow6_dll: str or Path
            Path to modflow6 .dll. You can obtain this library by downloading
            `the last iMOD5 release
            <https://oss.deltares.nl/web/imod/download-imod5>`_
        metaswap_dll: str or Path
            Path to metaswap .dll. You can obtain this library by downloading
            `the last iMOD5 release
            <https://oss.deltares.nl/web/imod/download-imod5>`_
        metaswap_dll_dependency: str or Path
            Directory with metaswap .dll dependencies. Directory should contain:
            [fmpich2.dll, mpich2mpi.dll, mpich2nemesis.dll, TRANSOL.dll]. You
            can obtain these by downloading `the last iMOD5 release
            <https://oss.deltares.nl/web/imod/download-imod5>`_
        ribasim_dll: str or Path
            Path to ribasim .dll.
        ribasim_dll_dependency: str or Path
            Directory with ribasim .dll dependencies.
        modflow6_write_kwargs: dict
            Optional dictionary with keyword arguments for the writing of
            Modflow6 models. You can use this for example to turn off the
            validation at writing (``validation=False``) or to write text files
            (``binary=False``)
        """

        if modflow6_write_kwargs is None:
            modflow6_write_kwargs = {}

        # force to Path
        directory = Path(directory)
        self.mf6_simulation.write(
            directory / self._modflow6_model_dir,
            **modflow6_write_kwargs,
        )
        self.msw_model.write(directory / self._metaswap_model_dir)
        self.ribasim_model.write(directory / self._ribasim_model_dir / "ribasim.toml")

        # Write exchange files
        exchange_dir = directory / "exchanges"
        exchange_dir.mkdir(mode=755, exist_ok=True)
        coupling_dict = self._get_coupling_dict(
            exchange_dir, self.mf6_rch_pkgkey, self.mf6_wel_pkgkey
        )
        self.write_exchanges(
            exchange_dir,
            self.mf6_rch_pkgkey,
            self.mf6_wel_pkgkey,
            coupling_dict,
        )
        self.write_toml(
            directory,
            coupling_dict,
            modflow6_dll,
            metaswap_dll,
            metaswap_dll_dependency,
            ribasim_dll,
            ribasim_dll_dependency,
        )

    def write_toml(
        self,
        directory: str | Path,
        coupling_dict: dict[str, Any],
        modflow6_dll: str | Path,
        metaswap_dll: str | Path,
        metaswap_dll_dependency: str | Path,
        ribasim_dll: str | Path,
        ribasim_dll_dependency: str | Path,
    ) -> None:
        """
        Write .toml file which configures the imod coupler run.

        Parameters
        ----------
        directory: str or Path
            Directory in which to write the .toml file.
        modflow6_dll: str or Path
            Path to modflow6 .dll. You can obtain this library by downloading
            `the last iMOD5 release
            <https://oss.deltares.nl/web/imod/download-imod5>`_
        metaswap_dll: str or Path
            Path to metaswap .dll. You can obtain this library by downloading
            `the last iMOD5 release
            <https://oss.deltares.nl/web/imod/download-imod5>`_
        metaswap_dll_dependency: str or Path
            Directory with metaswap .dll dependencies. Directory should contain:
            [fmpich2.dll, mpich2mpi.dll, mpich2nemesis.dll, TRANSOL.dll]. You
            can obtain these by downloading `the last iMOD5 release
            <https://oss.deltares.nl/web/imod/download-imod5>`_
        ribasim_dll: str or Path
            Path to ribasim .dll.
        ribasim_dll_dependency: str or Path
            Directory with ribasim .dll dependencies.
        """
        # force to Path
        directory = Path(directory)
        directory.mkdir(parents=True, exist_ok=True)

        toml_path = directory / self._toml_name
        coupler_toml = {
            "timing": False,
            "log_level": "INFO",
            "driver_type": "ribametamod",
            "driver": {
                "kernels": {
                    "modflow6": {
                        "dll": str(modflow6_dll),
                        "work_dir": self._modflow6_model_dir,
                    },
                    "metaswap": {
                        "dll": str(metaswap_dll),
                        "dll_dep_dir": str(metaswap_dll_dependency),
                        "work_dir": self._metaswap_model_dir,
                    },
                    "ribasim": {
                        "dll": str(ribasim_dll),
                        "dll_dep_dir": str(ribasim_dll_dependency),
                        "config_file": str(
                            Path(self._ribasim_model_dir) / "ribasim.toml"
                        ),
                    },
                },
                "coupling": [coupling_dict],
            },
        }

        with open(toml_path, "wb") as f:
            tomli_w.dump(coupler_toml, f)

    @staticmethod
    def validate_keys(
        gwf_model: GroundwaterFlowModel,
        active_keys: list[str],
        passive_keys: list[str],
        expected_type: River | Drainage,
    ) -> None:
        active_keys_set = set(active_keys)
        passive_keys_set = set(passive_keys)
        intersection = active_keys_set.intersection(passive_keys_set)
        if intersection:
            raise ValueError(f"active and passive keys share members: {intersection}")
        present = [k for k, v in gwf_model.items() if isinstance(v, expected_type)]
        missing = (active_keys_set | passive_keys_set).difference(present)
        if missing:
            raise ValueError(
                f"keys with expected type {expected_type.__name__} are not "
                f"present in the model: {missing}"
            )

    @staticmethod
    def derive_river_drainage_coupling(
        gridded_basin_mod: xr.DataArray,
        basin_ids: pd.Series,
        conductance: xr.DataArray,
    ) -> pd.DataFrame:
        # Conductance is leading parameter to define location, for both river
        # and drainage.
        # FUTURE: check for time dimension? Also order and inclusion of layer
        # in conductance.
        # Use xarray where to force the dimension order of conductance.
        basin_id = xr.where(conductance.notnull(), gridded_basin_mod, np.nan)  # type: ignore
        include = basin_id.notnull().to_numpy()
        basin_id_values = basin_id.to_numpy()[include].astype(int)
        # Ribasim internally sorts the basin, which determines the order of the
        # Ribasim level array.
        basin_index = np.searchsorted(basin_ids, basin_id_values)
        boundary_index_values = np.cumsum(conductance.notnull().to_numpy().ravel()) - 1
        boundary_index_values = boundary_index_values[include.ravel()]
        return pd.DataFrame(
            data={"basin_index": basin_index, "bound_index": boundary_index_values}
        )

    @staticmethod
    def derive_coupling(
        gridded_basin: xr.DataArray,
        basin_ids: "pd.Series[int]",
        condition: xr.DataArray,
    ) -> pd.DataFrame:
        # condition AND gridded_basin notnan are leading directive to define location which to couple to
        basin_id = xr.where(condition, gridded_basin, np.nan)  # type: ignore
        include = basin_id.notnull().to_numpy()
        basin_id_values = basin_id.to_numpy()[include].astype(int)
        basin_index = np.searchsorted(basin_ids, basin_id_values)
        boundary_index_values = np.arange(np.size(basin_index))
        return pd.DataFrame(
            data={"basin_index": basin_index, "bound_index": boundary_index_values}
        )

    def _get_coupling_dict(
        self,
        directory: Path,
        mf6_rch_pkgkey: str,
        mf6_wel_pkgkey: str | None,
    ) -> dict[str, Any]:
        """
        Get dictionary with names of coupler packages and paths to mappings.

        Parameters
        ----------
        directory: Path
            Directory where .dxc files are written.
        mf6_rch_pkgkey: str
            Key of Modflow 6 recharge package to which MetaSWAP is coupled.
        mf6_wel_pkgkey: str
            Key of Modflow 6 well package to which MetaSWAP sprinkling is
            coupled.

        Returns
        -------
        coupling_dict: dict
            Dictionary with names of coupler packages and paths to mappings.
        """

        coupling_dict: dict[str, Any] = {}

        gwf_names = self._get_gwf_modelnames()

        # Assume only one groundwater flow model
        # FUTURE: Support multiple groundwater flow models.
        coupling_dict["mf6_model"] = gwf_names[0]

        coupling_dict[
            "mf6_msw_node_map"
        ] = f"./{directory.name}/{NodeSvatMapping._file_name}"

        coupling_dict["mf6_msw_recharge_pkg"] = mf6_rch_pkgkey
        coupling_dict[
            "mf6_msw_recharge_map"
        ] = f"./{directory.name}/{RechargeSvatMapping._file_name}"

        coupling_dict["enable_sprinkling"] = self.is_sprinkling

        if self.is_sprinkling:
            coupling_dict["mf6_msw_well_pkg"] = mf6_wel_pkgkey
            coupling_dict[
                "mf6_msw_sprinkling_map"
            ] = f"./{directory.name}/{WellSvatMapping._file_name}"

        return coupling_dict

    def write_exchanges(
        self,
        directory: str | Path,
        mf6_rch_pkgkey: str,
        mf6_wel_pkgkey: str | None,
        coupling_dict: dict[str, Any],
    ) -> None:
        gwf_names = self._get_gwf_modelnames()

        # force to Path
        directory = Path(directory)

        # Assume only one groundwater flow model
        # FUTURE: Support multiple groundwater flow models.
        gwf_model = self.mf6_simulation[gwf_names[0]]
        coupling = self.coupling_list[0]
        self.validate_keys(
            gwf_model,
            coupling.mf6_active_river_packages,
            coupling.mf6_passive_river_packages,
            River,
        )
        self.validate_keys(
            gwf_model,
            coupling.mf6_active_drainage_packages,
            coupling.mf6_passive_drainage_packages,
            Drainage,
        )

        dis = gwf_model[gwf_model._get_pkgkey("dis")]
        # gridded basin riba with respect to the MODFLOW grid
        gridded_basin_mod = imod.prepare.rasterize(
            self.basin_definition,
            like=dis["idomain"].isel(layer=0, drop=True),
            column="node_id",
        )

        grid_data_key = [
            pkgname
            for pkgname, pkg in self.msw_model.items()
            if isinstance(pkg, GridData)
        ][0]
        index, svat = self.msw_model[grid_data_key].generate_index_array()
        grid_mapping = NodeSvatMapping(svat, dis)
        grid_mapping.write(directory, index, svat)

        recharge = gwf_model[mf6_rch_pkgkey]
        rch_mapping = RechargeSvatMapping(svat, recharge)
        rch_mapping.write(directory, index, svat)

        # mapping ponding: metaswap - ribasim
        # gridded basin riba with respect to the MetaSwap svat grid
        gridded_basin_msw = imod.prepare.rasterize(
            self.basin_definition,
            like=svat.isel(subunit=0, drop=True),
            column="node_id",
        )

        assert self.ribasim_model.basin.profile.df is not None
        basin_ids = np.unique(self.ribasim_model.basin.profile.df["node_id"])
        missing = ~np.isin(self.basin_definition["node_id"], basin_ids)
        if missing.any():
            missing_basins = self.basin_definition["node_id"].to_numpy()[missing]
            raise ValueError(
                "The node IDs of these basins in the basin definition do not "
                f"occur in the Ribasim model: {missing_basins}"
            )

        packages = asdict(coupling)
        for destination in packages:
            coupling_dict[destination] = {}
        coupling_dict["mf6_model"] = packages.pop("mf6_model")
        for destination, keys in packages.items():
            for key in keys:
                package = gwf_model[key]
                table = self.derive_river_drainage_coupling(
                    gridded_basin_mod, basin_ids, package["conductance"]
                )
                table.to_csv(directory / f"{key}.tsv", sep="\t", index=False)
                coupling_dict[destination][key] = f"exchanges/{key}.tsv"

        # ponding for all svats
        table_ponding = self.derive_coupling(gridded_basin_msw, basin_ids, svat > 0)
        table_ponding.to_csv(directory / "msw_ponding.tsv", sep="\t", index=False)
        coupling_dict["rib_msw_ponding_map_surface_water"] = "exchanges/msw_ponding.tsv"

        if self.is_sprinkling:
            # sprinkling groundwater
            well = gwf_model[mf6_wel_pkgkey]
            well_mapping = WellSvatMapping(svat, well)
            well_mapping.write(directory, index, svat)

            # sprinkling surface water for subsection of svats determined in 'sprinkling'
            swspr_grid_data = copy.deepcopy(self.msw_model[grid_data_key])
            nsu = swspr_grid_data.dataset["area"].sizes["subunit"]
            swsprmax = self.msw_model["sprinkling"]
            swspr_grid_data.dataset["area"].values = np.tile(
                swsprmax["max_abstraction_surfacewater_m3_d"].values,
                (nsu, 1, 1),
            )
            _, svat_swspr = swspr_grid_data.generate_index_array()
            table_sw_sprinkling = self.derive_coupling(
                gridded_basin_msw, basin_ids, svat_swspr.notnull()
            )
            table_sw_sprinkling.to_csv(
                directory / "msw_sw_sprinkling.tsv", sep="\t", index=False
            )
            coupling_dict[
                "rib_msw_sprinkling_map_surface_water"
            ] = "exchanges/msw_sw_sprinkling.tsv"



from primod.mapping.node_svat_mapping import NodeSvatMapping
from primod.mapping.rch_svat_mapping import RechargeSvatMapping
from primod.mapping.wel_svat_mapping import WellSvatMapping
from primod.ribametamod.ribametamod import DriverCoupling, RibaMetaMod

__all__ = [
    "RibaMetaMod",
    "DriverCoupling",
    "NodeSvatMapping",
    "RechargeSvatMapping",
    "WellSvatMapping",
]



import numpy as np
import pandas as pd
import xarray as xr
from numpy.typing import NDArray
from scipy.spatial import KDTree

from primod.typing import Bool, Float, Int


def _ensure_time_invariant_conductance(conductance: xr.DataArray) -> xr.DataArray:
    # Check that the conductance location does not change over time.
    # The imod_coupler has static weights: changing locations means having
    # to update the coupling weights over time as well.
    if "time" in conductance.dims:
        notnull = conductance.notnull()
        if (notnull.any("time") != notnull.all("time")).any():
            raise ValueError(
                "For imod_coupler, the active cells (defined by the conductance)"
                "in a river or drainage package must be constant over time."
            )
        return conductance.isel(time=0, drop=True)
    else:
        return conductance


def _find_coupled_cells(
    conductance: xr.DataArray,
    gridded_basin: xr.DataArray,
    basin_ids: pd.Series,
) -> tuple[NDArray[Int], NDArray[Bool]]:
    """
    Compare the location of conductance with the gridded basin.

    Either may have values where the other has NaN values.
    Only cells that have values in the conductance AND in the gridded
    basin should be coupled.
    """
    basin_id = xr.where(conductance.notnull(), gridded_basin, np.nan)  # type: ignore
    include = basin_id.notnull().to_numpy()
    basin_id_values = basin_id.to_numpy()[include].astype(int)
    # Ribasim internally sorts the basin, which determines the order of the
    # Ribasim state arrays.
    basin_index = np.searchsorted(basin_ids, basin_id_values)
    return basin_index, include


def _derive_boundary_index(
    conductance: xr.DataArray,
    include: NDArray[Bool],
) -> NDArray[Int]:
    """
    Create the numbering for each boundary cell.

    E.g. a 3 by 3 conductance grid:
    [
        [nan, 1.0, np.nan],
        [nan, 1.0, np.nan],
        [2.0, 1.0, 4.0],
    ]

    Will have essentially the boundary_index_values:

    [
        [None, 0, None],
        [None, 1, None],
        [2, 3, 4],
    ]

    However, the gridded_basin might look like:

    [
        [nan, 1, nan],
        [nan, 1, nan],
        [nan, 2, 2],
    ]

    I.e. the bottom-left cell is not included in the basin definition.

    Then, the desired return values are: [0, 1, 3, 4]

    This implements this in a vectorized manner.
    """
    # The first value found by cumsum will have a 1, while the coupler is 0-indexed.
    boundary_index_values = np.cumsum(conductance.notnull().to_numpy().ravel()) - 1
    selection: NDArray[Int] = boundary_index_values[include.ravel()]
    return selection


def derive_passive_coupling(
    conductance: xr.DataArray,
    gridded_basin: xr.DataArray,
    basin_ids: pd.Series,
) -> pd.DataFrame:
    # Conductance is leading parameter to define location, for both river
    # and drainage.
    # Use xarray.where() to force the dimension order of conductance, rather than
    # using gridded_basin.where() (which prioritizes the gridded_basin dims)
    conductance = _ensure_time_invariant_conductance(conductance)
    basin_index, include = _find_coupled_cells(conductance, gridded_basin, basin_ids)
    boundary_index = _derive_boundary_index(conductance, include)
    return pd.DataFrame(
        data={"basin_index": basin_index, "bound_index": boundary_index}
    )


def _get_subgrid_xy(subgrid: pd.DataFrame) -> NDArray[Float]:
    # Check whether columns (optional to Ribasim) are present.
    if "meta_x" not in subgrid or "meta_y" not in subgrid:
        raise ValueError(
            'The columns "meta_x" and "meta_y" are required in the '
            "ribasim.Model.basin.subgrid dataframe for actively coupled river "
            f"and drainage packages. Found columns: {subgrid.columns}"
        )

    # Check x and y coordinates for uniqueness
    grouped = subgrid.groupby("subgrid_id")
    multiple_x = grouped["meta_x"].nunique().ne(1)
    multiple_y = grouped["meta_y"].nunique().ne(1)
    if multiple_x.any() or multiple_y.any():
        x_ids = multiple_x.index[multiple_x].to_numpy()
        y_ids = multiple_y.index[multiple_y].to_numpy()
        raise ValueError(
            "Subgrid data contains multiple values for meta_x or meta_y "
            f"for subgrid_id(s):\n   meta_x: {x_ids}\n   meta_y: {y_ids}"
        )

    x = grouped["meta_x"].first().to_numpy()
    y = grouped["meta_y"].first().to_numpy()
    return np.column_stack((x, y))


def _get_conductance_xy(
    conductance: xr.DataArray, include: NDArray[Bool]
) -> NDArray[Float]:
    # Include incorporates the layer dimension
    # We don't want to replicate x and y for each layer.
    # Instead, compute the indices into the x and y array for a single layer.
    x = conductance["x"].to_numpy()
    y = conductance["y"].to_numpy()
    n_per_layer = y.size * x.size
    include2d = np.flatnonzero(include) % n_per_layer
    yy, xx = np.meshgrid(y, x, indexing="ij")
    xy = np.column_stack((xx.ravel(), yy.ravel()))
    conductance_xy: NDArray[Float] = xy[include2d]
    return conductance_xy


def _find_nearest_subgrid_elements(
    subgrid_xy: NDArray[Float],
    conductance_xy: NDArray[Float],
) -> NDArray[Int]:
    """Find the nearest subgrid element using a KD Tree."""
    kdtree = KDTree(subgrid_xy)
    # FUTURE: maybe do something with the distances returned by query?
    indices: NDArray[Int] = kdtree.query(conductance_xy)[1]
    return indices


def derive_active_coupling(
    conductance: xr.DataArray,
    gridded_basin: xr.DataArray,
    basin_ids: pd.Series,
    subgrid_df: pd.DataFrame,
) -> pd.DataFrame:
    # Conductance is leading parameter to define location, for both river
    # and drainage.
    # Use xarray.where() to force the dimension order of conductance, rather than
    # using gridded_basin.where() (which prioritizes the gridded_basin dims)
    conductance = _ensure_time_invariant_conductance(conductance)
    basin_index, include = _find_coupled_cells(conductance, gridded_basin, basin_ids)
    boundary_index = _derive_boundary_index(conductance, include)
    # Match the cells to the subgrid based on xy location.
    subgrid_xy = _get_subgrid_xy(subgrid_df)
    conductance_xy = _get_conductance_xy(conductance, include)
    subgrid_index = _find_nearest_subgrid_elements(subgrid_xy, conductance_xy)
    return pd.DataFrame(
        data={
            "basin_index": basin_index,
            "bound_index": boundary_index,
            "subgrid_index": subgrid_index,
        }
    )



import abc
from io import TextIOWrapper
from pathlib import Path
from typing import Any

import pandas as pd
import xarray as xr
from imod.msw.fixed_format import VariableMetaData, format_fixed_width
from numpy.typing import NDArray


class GenericMapping(abc.ABC):
    """
    MappingBase is used to share methods for specific packages with no time
    component for multiple two-component mappings.

    It is intended as an abstract base class, only to inherit from, to implement new
    packages.
    """

    __slots__ = "_pkg_id", "dataset"
    _metadata_dict: dict[str, VariableMetaData]
    _with_subunit: tuple[str, str, str]
    _to_fill: tuple[str]
    _file_name: str

    def __init__(self) -> None:
        self.dataset = xr.Dataset()

    def __getitem__(self, key: str) -> Any:
        return self.dataset.__getitem__(key)

    def __setitem__(self, key: str, value: Any) -> None:
        self.dataset.__setitem__(key, value)

    def isel(self) -> None:
        raise NotImplementedError(
            "Selection on packages not yet supported. "
            "To make a selection on the xr.Dataset, call dataset.isel instead. "
            "You can create a new package with a selection by calling (**dataset.isel(**selection))"
        )

    def sel(self) -> None:
        raise NotImplementedError(
            "Selection on packages not yet supported. "
            "To make a selection on the xr.Dataset, call dataset.sel instead. "
            "You can create a new package with a selection by calling (**dataset.sel(**selection))"
        )

    def _check_range(self, dataframe: pd.DataFrame) -> None:
        for varname in dataframe:
            min_value = self._metadata_dict[varname].min_value
            max_value = self._metadata_dict[varname].max_value
            if (dataframe[varname] < min_value).any() or (
                dataframe[varname] > max_value
            ).any():
                raise ValueError(
                    f"{varname}: not all values are within range ({min_value}-{max_value})."
                )

    def write_dataframe_fixed_width(
        self, file: TextIOWrapper, dataframe: pd.DataFrame
    ) -> None:
        for row in dataframe.itertuples():
            for index, metadata in enumerate(self._metadata_dict.values()):
                content = format_fixed_width(row[index + 1], metadata)
                file.write(content)
            file.write("\n")

    def _index_da(self, da: pd.DataFrame, index: NDArray[Any]) -> Any:
        return da.to_numpy().ravel()[index]

    def _render(
        self, file: TextIOWrapper, index: NDArray[Any], svat: pd.DataFrame
    ) -> None:
        data_dict = {"svat": svat.to_numpy().ravel()[index]}

        for var in self._with_subunit:
            data_dict[var] = self._index_da(self.dataset[var], index)

        for var in self._to_fill:
            data_dict[var] = ""

        dataframe = pd.DataFrame(
            data=data_dict, columns=list(self._metadata_dict.keys())
        )

        self._check_range(dataframe)
        self.write_dataframe_fixed_width(file, dataframe)

    def write(
        self, directory: str | Path, index: NDArray[Any], svat: xr.DataArray
    ) -> None:
        """
        Write mapping to .dxc file.

        Parameters
        ----------
        directory: str or Path
            directory in which exchange file should be written
        index: np.array

        """
        # Force to Path
        directory = Path(directory)

        filename = directory / self._file_name
        with open(filename, "w") as f:
            self._render(f, index, svat)


class MetaModMapping(GenericMapping):
    """
    MetaModMapping is used to share methods for specific packages with no time
    component.

    It is not meant to be used directly, only to inherit from, to implement new
    packages.
    """


class RibaMetaMapping(GenericMapping):
    """
    MetaRibaMapping is used to share methods for specific packages with no time
    component.

    It is not meant to be used directly, only to inherit from, to implement new
    packages.
    """



from primod.mapping.node_svat_mapping import NodeSvatMapping
from primod.mapping.rch_svat_mapping import RechargeSvatMapping
from primod.mapping.wel_svat_mapping import WellSvatMapping
from primod.metamod.metamod import MetaMod

__all__ = ["MetaMod", "NodeSvatMapping", "RechargeSvatMapping", "WellSvatMapping"]



from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import Any

import geopandas as gpd
import imod
import numpy as np
import pandas as pd
import ribasim
import tomli_w
import xarray as xr
from imod.mf6 import Drainage, GroundwaterFlowModel, Modflow6Simulation, River
from numpy.typing import NDArray

from primod.ribamod.exchange_creator import (
    derive_active_coupling,
    derive_passive_coupling,
)
from primod.typing import Int


@dataclass
class DriverCoupling:
    """A dataclass representing one coupling scenario for the RibaMod driver.

    Attributes
    ----------
    mf6_model : str
        The model of the driver.
    basin_definition : gpd.GeoDataFrame
        GeoDataFrame of basin polygons
    mf6_active_river_packages : list of str
        A list of active river packages.
    mf6_passive_river_packages : list of str
        A list of passive river packages.
    mf6_active_drainage_packages : list of str
        A list of active drainage packages.
    mf6_passive_drainage_packages : list of str
        A list of passive drainage packages.
    """

    mf6_model: str
    basin_definition: gpd.GeoDataFrame
    mf6_active_river_packages: list[str] = field(default_factory=list)
    mf6_passive_river_packages: list[str] = field(default_factory=list)
    mf6_active_drainage_packages: list[str] = field(default_factory=list)
    mf6_passive_drainage_packages: list[str] = field(default_factory=list)


class RibaMod:
    """Couple Ribasim and MODFLOW 6.

    Parameters
    ----------
    ribasim_model : ribasim.model
        The Ribasim model that should be coupled.
    mf6_simulation : Modflow6Simulation
        The Modflow6 simulation that should be coupled.
    coupling_list: list of DriverCoupling
        One entry per MODFLOW 6 model that should be coupled
    """

    _toml_name = "imod_coupler.toml"
    _ribasim_model_dir = "ribasim"
    _modflow6_model_dir = "modflow6"

    def __init__(
        self,
        ribasim_model: ribasim.Model,
        mf6_simulation: Modflow6Simulation,
        coupling_list: list[DriverCoupling],
    ):
        self.validate_time_window(
            ribasim_model=ribasim_model,
            mf6_simulation=mf6_simulation,
        )
        self.ribasim_model = ribasim_model
        self.mf6_simulation = mf6_simulation
        self.coupling_list = coupling_list

    def _get_gwf_modelnames(self) -> list[str]:
        """
        Get names of gwf models in mf6 simulation
        """
        return [
            key
            for key, value in self.mf6_simulation.items()
            if isinstance(value, GroundwaterFlowModel)
        ]

    @staticmethod
    def _empty_coupling_dict() -> dict[str, Any]:
        return {
            "mf6_active_river_packages": {},
            "mf6_passive_river_packages": {},
            "mf6_active_drainage_packages": {},
            "mf6_passive_drainage_packages": {},
        }

    def write(
        self,
        directory: str | Path,
        modflow6_dll: str | Path,
        ribasim_dll: str | Path,
        ribasim_dll_dependency: str | Path,
        modflow6_write_kwargs: dict[str, Any] | None = None,
    ) -> None:
        """
        Write Ribasim and Modflow 6 model with exchange files, as well as a
        ``.toml`` file which configures the iMOD Coupler run.

        Parameters
        ----------
        directory: str or Path
            Directory in which to write the coupled models
        modflow6_dll: str or Path
            Path to modflow6 .dll. You can obtain this library by downloading
            `the last iMOD5 release
            <https://oss.deltares.nl/web/imod/download-imod5>`_
        ribasim_dll: str or Path
            Path to ribasim .dll.
        ribasim_dll_dependency: str or Path
            Directory with ribasim .dll dependencies.
        modflow6_write_kwargs: dict
            Optional dictionary with keyword arguments for the writing of
            Modflow6 models. You can use this for example to turn off the
            validation at writing (``validation=False``) or to write text files
            (``binary=False``)
        """

        if modflow6_write_kwargs is None:
            modflow6_write_kwargs = {}

        # force to Path
        directory = Path(directory)
        coupling_dict, coupled_basins = self.write_exchanges(directory)

        self._nullify_ribasim_exchange_input(coupled_basins)

        self.mf6_simulation.write(
            directory / self._modflow6_model_dir,
            **modflow6_write_kwargs,
        )
        self.ribasim_model.write(directory / self._ribasim_model_dir / "ribasim.toml")
        self.write_toml(
            directory,
            coupling_dict,
            modflow6_dll,
            ribasim_dll,
            ribasim_dll_dependency,
        )

    def write_toml(
        self,
        directory: str | Path,
        coupling_dict: dict[str, Any],
        modflow6_dll: str | Path,
        ribasim_dll: str | Path,
        ribasim_dll_dependency: str | Path,
    ) -> None:
        """
        Write .toml file which configures the imod coupler run.

        Parameters
        ----------
        directory: str or Path
            Directory in which to write the .toml file.
        modflow6_dll: str or Path
            Path to modflow6 .dll. You can obtain this library by downloading
            `the last iMOD5 release
            <https://oss.deltares.nl/web/imod/download-imod5>`_
        ribasim_dll: str or Path
            Path to ribasim .dll.
        ribasim_dll_dependency: str or Path
            Directory with ribasim .dll dependencies.
        """
        # force to Path
        directory = Path(directory)
        directory.mkdir(parents=True, exist_ok=True)

        toml_path = directory / self._toml_name
        coupler_toml = {
            "timing": False,
            "log_level": "INFO",
            "driver_type": "ribamod",
            "driver": {
                "kernels": {
                    "modflow6": {
                        "dll": str(modflow6_dll),
                        "work_dir": self._modflow6_model_dir,
                    },
                    "ribasim": {
                        "dll": str(ribasim_dll),
                        "dll_dep_dir": str(ribasim_dll_dependency),
                        "config_file": str(
                            Path(self._ribasim_model_dir) / "ribasim.toml"
                        ),
                    },
                },
                "coupling": [coupling_dict],
            },
        }

        with open(toml_path, "wb") as f:
            tomli_w.dump(coupler_toml, f)

    @staticmethod
    def validate_keys(
        gwf_model: GroundwaterFlowModel,
        active_keys: list[str],
        passive_keys: list[str],
        expected_type: River | Drainage,
    ) -> None:
        active_keys_set = set(active_keys)
        passive_keys_set = set(passive_keys)
        intersection = active_keys_set.intersection(passive_keys_set)
        if intersection:
            raise ValueError(f"active and passive keys share members: {intersection}")
        present = [k for k, v in gwf_model.items() if isinstance(v, expected_type)]
        missing = (active_keys_set | passive_keys_set).difference(present)
        if missing:
            raise ValueError(
                f"keys with expected type {expected_type.__name__} are not "
                f"present in the model: {missing}"
            )

    @staticmethod
    def validate_time_window(
        ribasim_model: ribasim.Model,
        mf6_simulation: Modflow6Simulation,
    ) -> None:
        def to_timestamp(xr_time: xr.DataArray) -> pd.Timestamp:
            return pd.Timestamp(xr_time.to_numpy().item())

        mf6_timedis = mf6_simulation["time_discretization"].dataset
        mf6_start = to_timestamp(mf6_timedis["time"].isel(time=0)).to_pydatetime()
        time_delta = pd.to_timedelta(
            mf6_timedis["timestep_duration"].isel(time=-1).item(), unit="days"
        )
        mf6_end = (
            to_timestamp(mf6_timedis["time"].isel(time=-1)) + time_delta
        ).to_pydatetime()

        ribasim_start = ribasim_model.starttime
        ribasim_end = ribasim_model.endtime
        if ribasim_start != mf6_start or ribasim_end != mf6_end:
            raise ValueError(
                "Ribasim simulation time window does not match MODFLOW6.\n"
                f"Ribasim: {ribasim_start} to {ribasim_end}\n"
                f"MODFLOW6: {mf6_start} to {mf6_end}\n"
            )
        return

    def validate_basin_node_ids(self, basin_definition: gpd.GeoDataFrame) -> pd.Series:
        assert self.ribasim_model.basin.profile.df is not None
        basin_ids: NDArray[Int] = np.unique(
            self.ribasim_model.basin.profile.df["node_id"]
        )
        missing = ~np.isin(basin_definition["node_id"], basin_ids)
        if missing.any():
            missing_basins = basin_definition["node_id"].to_numpy()[missing]
            raise ValueError(
                "The node IDs of these basins in the basin definition do not "
                f"occur in the Ribasim model: {missing_basins}"
            )
        return basin_ids

    def validate_subgrid_df(self, coupling: DriverCoupling) -> pd.DataFrame | None:
        if coupling.mf6_active_river_packages or coupling.mf6_active_drainage_packages:
            if self.ribasim_model.basin.subgrid.df is None:
                raise ValueError(
                    "ribasim.model.basin.subgrid must be defined for actively coupled packages."
                )
            return self.ribasim_model.basin.subgrid.df
        else:
            return None

    def _nullify_ribasim_exchange_input(
        self, coupled_basin_node_ids: NDArray[Int]
    ) -> None:
        """
        Set the input forcing to NoData for drainage and infiltration.

        Ribasim will otherwise overwrite the values set by the coupler.
        """

        # FUTURE: in coupling to MetaSWAP, the runoff should be set nodata as well.
        def _nullify(df: pd.DataFrame) -> None:
            """Set drainage and infiltration columns to nodata if present in df"""
            if df is not None:
                columns_present = list(
                    {"drainage", "infiltration"}.intersection(df.columns)
                )
                if len(columns_present) > 0:
                    df.loc[
                        df["node_id"].isin(coupled_basin_node_ids), columns_present
                    ] = np.nan
            return

        basin = self.ribasim_model.basin
        _nullify(basin.static.df)
        _nullify(basin.time.df)
        return

    def _process_driver_coupling(
        self,
        gwf_model: GroundwaterFlowModel,
        coupling: DriverCoupling,
    ) -> tuple[dict[str, Any], list[NDArray[Int]]]:
        dis = gwf_model[gwf_model._get_pkgkey("dis")]
        packages = asdict(coupling)
        packages.pop("mf6_model")
        basin_definition = packages.pop("basin_definition")

        # Validate
        if "node_id" not in basin_definition.columns:
            raise ValueError(
                'Basin definition must contain "node_id" column.'
                f"Columns in dataframe: {basin_definition.columns}"
            )
        # Check whether names in driver_coupling are present in MF6 model.
        self.validate_keys(
            gwf_model,
            coupling.mf6_active_river_packages,
            coupling.mf6_passive_river_packages,
            River,
        )
        self.validate_keys(
            gwf_model,
            coupling.mf6_active_drainage_packages,
            coupling.mf6_passive_drainage_packages,
            Drainage,
        )

        # Spatial overlays of MF6 boundaries with basin polygons.
        basin_ids = self.validate_basin_node_ids(basin_definition)
        subgrid_df = self.validate_subgrid_df(coupling)
        gridded_basin = imod.prepare.rasterize(
            basin_definition,
            like=dis["idomain"].isel(layer=0, drop=True),
            column="node_id",
        )

        # Collect which Ribasim basins are coupled, so that we can set their
        # drainage and infiltration to NoData later on.
        tables_dict = self._empty_coupling_dict()
        coupled_basin_indices = []
        for destination, keys in packages.items():
            for key in keys:
                package = gwf_model[key]
                # Derive exchange tables
                if "active" in destination:
                    table = derive_active_coupling(
                        gridded_basin=gridded_basin,
                        basin_ids=basin_ids,
                        conductance=package["conductance"],
                        subgrid_df=subgrid_df,
                    )
                else:
                    table = derive_passive_coupling(
                        gridded_basin=gridded_basin,
                        basin_ids=basin_ids,
                        conductance=package["conductance"],
                    )

                if table.empty:
                    raise ValueError(
                        f"No coupling can be derived for MODFLOW 6 package: {key}. "
                        "No spatial overlap exists between the basin_definition and this package."
                    )

                tables_dict[destination][key] = table
                coupled_basin_indices.append(table["basin_index"])

        coupled_basin_indices = np.unique(np.concatenate(coupled_basin_indices))
        coupled_basin_node_ids = basin_ids[coupled_basin_indices]
        return tables_dict, coupled_basin_node_ids

    def write_exchanges(
        self,
        directory: str | Path,
    ) -> tuple[dict[str, dict[str, str]], NDArray[Int]]:
        # Assume only one groundwater flow model
        # FUTURE: Support multiple groundwater flow models.
        gwf_names = self._get_gwf_modelnames()
        gwf_model = self.mf6_simulation[gwf_names[0]]

        exchange_dir = Path(directory) / "exchanges"
        exchange_dir.mkdir(exist_ok=True, parents=True)

        coupled_node_indices = []
        list_of_tables_dicts = []
        mf6_models = []
        for coupling in self.coupling_list:
            mf6_models.append(coupling.mf6_model)
            tables_dict, basin_node_indices = self._process_driver_coupling(
                gwf_model=gwf_model,
                coupling=coupling,
            )
            coupled_node_indices.append(basin_node_indices)
            list_of_tables_dicts.append(tables_dict)

        # FUTURE: if we support multiple MF6 models, group them by name before
        # merging, and return a list of coupling_dicts.
        merged_coupling_dict = self._empty_coupling_dict()
        merged_coupling_dict["mf6_model"] = mf6_models[0]
        for coupling_dict in list_of_tables_dicts:
            for destination, tables in coupling_dict.items():
                for key, table in tables.items():
                    # Write exchange table to TSV file.
                    table.to_csv(exchange_dir / f"{key}.tsv", sep="\t", index=False)
                    # Store path to TSV file for config.
                    merged_coupling_dict[destination][key] = f"exchanges/{key}.tsv"

        coupled_basin_node_ids = np.unique(np.concatenate(coupled_node_indices))
        return merged_coupling_dict, coupled_basin_node_ids



import numpy as np
import xarray as xr
from imod import mf6
from imod.msw.fixed_format import VariableMetaData
from numpy.typing import NDArray
from primod.mapping.mappingbase import MetaModMapping


class RechargeSvatMapping(MetaModMapping):
    """
    This contains the data to connect MODFLOW 6 recharge cells to MetaSWAP
    svats.

    This class is responsible for the file `rchindex2svat.dxc`.

    Parameters
    ----------
    svat: array of floats (xr.DataArray)
        SVAT units. This array must have a subunit coordinate to describe
        different land uses.
    recharge: mf6.Recharge
        Modflow 6 Recharge package to connect to. Note that the recharge rate
        should be provided as a 2D grid with a (y, x) dimension. MetaModMapping
        will throw an error if a grid is provided with different dimensions.
    """

    _file_name = "rchindex2svat.dxc"
    _metadata_dict = {
        "rch_id": VariableMetaData(10, 1, 9999999, int),
        "free": VariableMetaData(2, None, None, str),
        "svat": VariableMetaData(10, 1, 9999999, int),
        "layer": VariableMetaData(5, 0, 9999, int),
    }

    _with_subunit = ("rch_id", "svat", "layer")
    _to_fill = ("free",)

    def __init__(self, svat: xr.DataArray, recharge: mf6.Recharge):
        super().__init__()
        self.dataset["svat"] = svat
        self.dataset["layer"] = xr.full_like(svat, 1)
        rate = recharge.dataset["rate"]
        if "layer" in rate.dims:
            rate_no_layer = rate.squeeze("layer", drop=True)
        else:
            # 'layer' can be still in a coord, so ensure is also dropped
            rate_no_layer = rate.drop_vars("layer", errors="ignore")
        self.dataset["rch_active"] = rate_no_layer.notnull()
        self._pkgcheck()
        self._create_rch_id()

    def _create_rch_id(self) -> None:
        self.dataset["rch_id"] = xr.full_like(
            self.dataset["svat"], fill_value=0, dtype=np.int64
        )

        n_subunit = self.dataset["subunit"].size
        n_rch = self.dataset["rch_active"].sum()

        rch_active = self.dataset["rch_active"].to_numpy()

        # recharge does not have a subunit dimension, so tile for n_subunits
        rch_id: NDArray[np.int_] = np.tile(np.arange(1, n_rch + 1), (n_subunit, 1))

        self.dataset["rch_id"].to_numpy()[:, rch_active] = rch_id

    def _pkgcheck(self) -> None:
        rch_dims = self.dataset["rch_active"].dims
        if rch_dims != ("y", "x"):
            raise ValueError(
                "Recharge grid can only have dimensions ('y', 'x'). Got "
                f"{rch_dims} instead"
            )

        # Check if active msw cell inactive in recharge
        active = self.dataset["svat"] != 0
        inactive_in_rch = active > self.dataset["rch_active"]

        if inactive_in_rch.any():
            raise ValueError(
                "Active MetaSWAP cell detected in inactive cell in Modflow6 recharge"
            )



from io import TextIOWrapper
from typing import Any

import numpy as np
import pandas as pd
import xarray as xr
from imod import mf6
from imod.msw.fixed_format import VariableMetaData
from numpy.typing import NDArray
from primod.mapping.mappingbase import MetaModMapping


class WellSvatMapping(MetaModMapping):
    """
    This contains the data to connect MODFLOW 6 well cells to MetaSWAP svats.

    This class is responsible for the file `wellindex2svat.dxc`.

    Parameters
    ----------
    svat: array of floats (xr.DataArray)
        SVAT units. This array must have a subunit coordinate to describe
        different land uses.
    well: mf6.Well
        Modflow 6 Well package to connect to.
    """

    _file_name = "wellindex2svat.dxc"
    _metadata_dict = {
        "wel_id": VariableMetaData(10, 1, 9999999, int),
        "free": VariableMetaData(2, None, None, str),
        "svat": VariableMetaData(10, 1, 9999999, int),
        "layer": VariableMetaData(5, 0, 9999, int),
    }

    _with_subunit = ("wel_id", "svat", "layer")
    _to_fill = ("free",)

    def __init__(self, svat: xr.DataArray, well: mf6.WellDisStructured):
        super().__init__()
        self.well = well
        well_mod_id, well_svat, layer = self._create_well_id(svat)
        self.dataset["wel_id"] = well_mod_id
        self.dataset["svat"] = well_svat
        self.dataset["layer"] = layer

    def _create_well_id(
        self, svat: pd.DataFrame
    ) -> tuple[NDArray[Any], NDArray[Any], NDArray[Any]]:
        """
        Get modflow indices, svats, and layer number for the wells
        """
        # Convert to Python's 0-based index
        well_row = self.well["row"] - 1
        well_column = self.well["column"] - 1
        well_layer = self.well["layer"]

        n_subunit = svat["subunit"].size

        well_svat = svat.to_numpy()[:, well_row, well_column]
        well_active = well_svat != 0

        well_svat_1d = well_svat[well_active]

        # Tile well_layers for each subunit
        layer = np.tile(well_layer, (n_subunit, 1))
        layer_1d = layer[well_active]

        well_id = self.well.dataset.coords["index"] + 1
        well_id_1d = np.tile(well_id, (n_subunit, 1))[well_active]

        return (well_id_1d, well_svat_1d, layer_1d)

    def _render(self, file: TextIOWrapper, *args: Any) -> None:
        data_dict: dict[str, Any] = {}
        data_dict["svat"] = self.dataset["svat"].to_numpy()
        data_dict["layer"] = self.dataset["layer"].to_numpy()
        data_dict["wel_id"] = self.dataset["wel_id"].to_numpy()

        for var in self._to_fill:
            data_dict[var] = ""

        dataframe = pd.DataFrame(
            data=data_dict, columns=list(self._metadata_dict.keys())
        )

        self._check_range(dataframe)
        self.write_dataframe_fixed_width(file, dataframe)
